import logging
import os
from typing import List, Tuple, Dict, Union, Any
import glob
import numpy as np
from pydicom import dcmread, FileDataset, DataElement
from pydicom.tag import Tag


class SpinalScan:
    def __init__(
        self,
        volume: np.array,
        pixel_spacing: Union[np.array, list],
        slice_thickness: Union[float, int],
    ) -> None:
        """
        Initialize general class for a spinal scan to be used be SpineNet.

        Parameters
        ----------
        volume : np.array
            The volume of the scan voxels (of the orientation height x width x sagittal slices).
        pixel_spacing : Union[np.array, list]
            The pixel spacing of the scan slices in mm. The order is height, width.
        slice_thickness : Union[float, int]
            The distance between consecutive slices in mm.
        """
        self.volume = volume
        self.pixel_spacing = pixel_spacing
        self.slice_thickness = slice_thickness


def load_dicoms(
        paths: List[Union[os.PathLike, str, bytes]],
        require_extensions: bool = True,
        metadata_overwrites: dict = {},
    ) -> SpinalScan:
        '''
        Generate SpinalScan from paths to each DICOM slice,
        with checks such as ensure that the correct tags are present
        and that scan is sagittal

        Parameters
        ----------
        paths : List[Union[os.PathLike, str, bytes]]
            list of paths to DICOM files
        require_extensions : bool
            flag to require that all DICOM files in the `paths` list have the same `.dcm` extension
        metadata_overwrites : dict
            dictionary of metadata to overwrite in the scan. This can be PixelSpacing, SliceThickness and ImageOrientationPatient (which should be sagittal)

        Returns
        -------
        SpinalScan
            Object representing scan from the DICOM files
        '''

        if require_extensions:
            assert all(
                [".dcm" == path[-4:] for path in paths]
            ), "All paths must have .dcm extension. To ignore extension, set require_extensions=False"

        dicom_files = [dcmread(path) for path in paths]

        for idx, dicom_file in enumerate(dicom_files):
            dicom_files[idx] = overwrite_tags(dicom_file, metadata_overwrites)

        # check relevant tags are present and slices are all sagittal
        for dicom_idx, dicom_file in enumerate(dicom_files):
            missing_tags = check_missing_tags(dicom_file)
            if len(missing_tags) > 0:
                raise ValueError(
                    f"Missing tags in file {paths[dicom_idx]}: {missing_tags}"
                )
            is_sagittal = is_sagittal_dicom_slice(dicom_file)
            if not is_sagittal:
                dicom_files.pop(dicom_idx)
                #raise ValueError(
                logging.warning( f"File at {paths[dicom_idx]} is not a sagittal dicom slice")
        # sort slices by sagittal position
        dicom_files = sorted(
            dicom_files, key=lambda dicom_file: dicom_file.InstanceNumber
        )


        pixel_spacing = np.mean(
            [np.array(dicom_file.PixelSpacing) for dicom_file in dicom_files]
        , axis=0)
        slice_thickness = np.mean(
            [np.array(dicom_file.SliceThickness) for dicom_file in dicom_files]
        )
        volume = np.stack(
            [np.array(dicom_file.pixel_array) for dicom_file in dicom_files], axis=-1
        )

        return SpinalScan(
            volume=volume, pixel_spacing=pixel_spacing, slice_thickness=slice_thickness
        )


def is_sagittal_dicom_slice(dicom_file: FileDataset) -> bool:
    '''
    Check if a dicom slice is sagittal.

    Parameters
    ----------
    dicom_file : FileDataset
        dicom file to check

    Returns
    -------
    bool
        True if sagittal, False otherwise
    '''
    if Tag("ImageOrientationPatient") in dicom_file:
        image_orientation = np.array(dicom_file.ImageOrientationPatient).round()
        if (image_orientation[[0, 3]] == [0, 0]).all():
            return True
        else:
            return False
    else:
        raise ValueError("ImageOrientationPatient metadata not found in dicom file")


def overwrite_tags(dicom_file: FileDataset, metadata_overwrites: dict) -> FileDataset:
    '''
    Overwrite tags in a dicom file. Currently only overwrites PixelSpacing, SliceThickness and ImageOrientationPatient.

    Parameters
    ----------
    dicom_file : FileDataset
        dicom file to overwrite values in
    metadata_overwrites : dict
        dictionary of metadata to overwrite in the scan. This can be PixelSpacing, SliceThickness and ImageOrientationPatient (which should be sagittal)

    Returns
    -------
    FileDataset
        dicom file with overwritten metadata
    '''

    possible_overwrites = {
        "PixelSpacing": "DS",
        "SliceThickness": "DS",
        "ImageOrientationPatient": "DS",
    }

    for tag, value in metadata_overwrites.items():
        if tag not in possible_overwrites:
            raise NotImplementedError(f"Overwriting tag {tag} is not supported")
        else:
            if Tag(tag) in dicom_file:
                dicom_file[Tag(tag)] = DataElement(
                    Tag(tag), possible_overwrites[tag], value
                )
            else:
                dicom_file.add_new(Tag(tag), possible_overwrites[tag], value)
    return dicom_file


def check_missing_tags(dicom_file: FileDataset) -> List[str]:
    '''
    Find which tags are missing in a dicom file from PixelData, PixelSpacing, SliceThickness and ImageOrientationPatient (all are required by SpineNet).

    Parameters
    ----------
    dicom_file : FileDataset
        dicom file to check

    Returns
    -------
    List[str]
        list of missing tags
    '''

    required_tags = ["PixelData", "PixelSpacing", "SliceThickness", "InstanceNumber"]
    missing_tags = [
        tag_name for tag_name in required_tags if Tag(tag_name) not in dicom_file
    ]
    return missing_tags


def is_dicom_file(path: Union[os.PathLike, str, bytes]) -> bool:
    '''
    Check if a file is a dicom file.
    Parameters
    ----------
    path : Union[os.PathLike, str, bytes]
        path to file to check
    Returns
    -------
    bool
        True if dicom file, False otherwise
    '''
    try:
        dcmread(path)
        return True
    except:
        return False


def load_dicoms_from_folder(
    path: Union[os.PathLike, str, bytes],
    require_extensions: bool = True,
    metadata_overwrites: dict = {},
) -> SpinalScan:
    '''
    Load a DICOM scan from a folder containing each of the slices.

    Parameters
    ----------
    path : Union[os.PathLike, str, bytes]
        path to folder containing dicom slices
    require_extensions : bool
        if True, requires all dicom files in the folder must have the extension .dcm.

    Returns
    -------
    SpinalScan
        Object representing scan from the DICOM files
    '''
    slices = [f for f in glob.glob(os.path.join(path, "*")) if is_dicom_file(f)]

    return load_dicoms(slices, require_extensions, metadata_overwrites)



import numpy as np


def format_gt_scores(score):
    if score == "":
        return None
    for key in score:
        score[key] = np.array(list(reversed(score[key])))
    formatted_gt_scores = {}
    formatted_gt_scores["Pfirrmann"] = label_check(score["Pfirrmann"]) - 1
    formatted_gt_scores["Narrowing"] = label_check(score["Narrowing"])
    formatted_gt_scores["UpperEndplateDefect"] = label_binarize(
        score["UpperEndplateDefect"]
    )
    formatted_gt_scores["LowerEndplateDefect"] = label_binarize(
        score["LowerEndplateDefect"]
    )
    formatted_gt_scores["UpperMarrow"] = label_check_marrow(
        score["UpperModic1"],
        score["UpperModic2"],
        score["UpperModic3"],
        score["UpperModicM"],
    )
    formatted_gt_scores["LowerMarrow"] = label_check_marrow(
        score["LowerModic1"],
        score["LowerModic2"],
        score["LowerModic3"],
        score["LowerModicM"],
    )
    formatted_gt_scores["Spondylolisthesis"] = label_check_3(score["Spondylolisthesis"])
    formatted_gt_scores["CentralCanalStenosis"] = label_check(
        score["CentralCanalStenosis"]
    )
    formatted_gt_scores["ForaminalStenosisLeft"] = label_binarize(
        score["ForaminalStenosisLeft"]
    )
    formatted_gt_scores["ForaminalStenosisRight"] = label_binarize(
        score["ForaminalStenosisRight"]
    )
    formatted_gt_scores["Herniation"] = label_binarize(score["Herniation"])
    for key in formatted_gt_scores:
        # return none if scores contain nan
        if (np.sum(formatted_gt_scores[key]) != np.sum(formatted_gt_scores[key])).any():
            return None
    return formatted_gt_scores


def label_binarize(label):
    label = label_check(label)
    label[label > 0] = 1

    return label


def label_check(label):
    label[np.isnan(label)] = -100
    return label


def label_check_marrow(m1, m2, m3, mm):
    m = m1 + m2 + m3 + mm
    m = label_binarize(m)
    return m


def label_check_3(label):
    label = label_check(label)
    label[label > 1] = 2
    return label



from typing import Dict, List, Tuple
from csv import DictWriter
import pandas as pd
VertDicts = List[Dict]

def save_vert_dicts_to_csv(vert_dicts: VertDicts, filename: str) -> None:
    '''
    Saves a list of vert_dicts to a file in csv format.

    Parameters
    ----------
    vert_dicts: List[Dict]
        A list of dictionaries with the following keys: 
        'polys', 'average_polygon', 'slice_nos', 'predicted_label'
        Each entry corresponds to a detected vertebral body
    filename: str
        The name of the file to save the vert_dicts to.
    '''

    required_keys = ['polys', 'average_polygon', 'slice_nos', 'predicted_label']
    check_no_keys_missing(vert_dicts, required_keys)

    with open(filename, 'w') as f:
        w = DictWriter(f, fieldnames=required_keys)
        w.writeheader()
        for vert_dict in vert_dicts:
            w.writerow(vert_dict)
    return


def check_no_keys_missing(vert_dicts, required_keys):
    for vert_dict in vert_dicts:
        for key in required_keys:
            if key not in vert_dict:
                raise KeyError(f'{key} is missing from vert_dict')
    



import torch
import numpy as np
import matplotlib.pyplot as plt

vert_names = [
    "S1",
    "L5",
    "L4",
    "L3",
    "L2",
    "L1",
    "T12",
    "T11",
    "T10",
    "T9",
    "T8",
    "T7",
    "T6",
    "T5",
    "T4",
    "T3",
    "T2",
    "T1",
    "C7",
    "C6",
    "C5",
    "C4",
    "C3",
    "C2",
]


def label_verts(
    vert_dicts,
    scan,
    pixel_spacing,
    appearance_net,
    context_net,
    plot_outputs=False,
    penalise_skips=False,
    debug=False,
):
    """
    Run the labelling pipeline to extract labels from each detection

    Parameters:
    -----------
    vert_dicts : list of dictionaries
        The output from the detection stage. Each dictionary contains information about each vertebra detected.
    scan : np.array
        HxWxS image containing the scan
    pixel_spacing : float
        the mm between each pixel in each sagittal slice (assumes isotropic in slice)
    appearance_net : torch.nn.Module
        the 3D apperaance network to guess a vert level from its surrounding area alone
    context_net : torch.nn.Module
        the context network used to refine predictions of the appearance_network
    plot_outputs : bool
        set to true to show guesses of appearance and refine ment network in a plot
    debug: whether to return information about the network outputs to see what
        went wrong


    Outputs
    -------
    vert_dicts : list of dictionaries
        same as the input to function but with a new key, predicted label added
    """

    # firstly get the labels from the apperance net
    vert_dicts = conv_appearance_labelling(vert_dicts, appearance_net, softmax_temp=10)
    # now construct input to the context model
    height_scaled_appearance_features, y_centroids = construct_input_to_context_model(
        vert_dicts, scan, pixel_spacing
    )
    # pad image to make it a multiple of 256 in height so it can be scaled down by U-Net
    diff = height_scaled_appearance_features.shape[0] % 256
    if diff != 0:
        height_scaled_appearance_features = np.concatenate(
            (height_scaled_appearance_features, np.zeros((256 - diff, 24))), axis=0
        )

    # run context network
    context_net_device = next(context_net.parameters()).device
    with torch.no_grad():
        context_output = context_net(
            torch.tensor(height_scaled_appearance_features)
            .float()
            .unsqueeze(0)
            .unsqueeze(0)
            .to(context_net_device)
        )

    context_features = [
        context_output[0, 0, int(y_centroid), :].tolist() for y_centroid in y_centroids
    ]
    for idx, vert_dict in enumerate(vert_dicts):
        vert_dict["context_features"] = context_features[idx]
        vert_dict["appearance_guess"] = vert_names[
            torch.argmax(vert_dict["visual_features"]).item()
        ]
        vert_dict["context_guess"] = vert_names[
            np.argmax(vert_dict["context_features"])
        ]

    # apply softmax with temp 10 to the output
    temperature = 10
    vert_centre_features = [
        torch.softmax(
            context_output[0, 0, int(y_centroid), :] / temperature, dim=0
        ).tolist()
        for y_centroid in y_centroids
    ]

    # perform beam search
    (
        sequence_predictions,
        same_both_directions,
        best_up_sequence,
        best_down_sequence,
    ) = two_way_beam_search(vert_centre_features, 100, penalise_skips=penalise_skips)
    # sequence_predictions = sequence_predictions[0]
    vert_centre_predictions = sequence_predictions[0][0]
    for idx, vert_dict in enumerate(vert_dicts):
        vert_dict["predicted_label"] = vert_names[vert_centre_predictions[idx]]

    # label S2 if we have 2 S1s
    if len(vert_dicts) > 1:
        if (
            vert_dicts[-1]["predicted_label"] == "S1"
            and vert_dicts[-2]["predicted_label"] == "S1"
        ):
            vert_dicts[-1]["predicted_label"] = "S2"

        if (
            vert_dicts[0]["predicted_label"] == "S1"
            and vert_dicts[1]["predicted_label"] == "S1"
        ):
            vert_dicts[0]["predicted_label"] = "S2"

    if plot_outputs:
        plt.figure(3, figsize=(15, 15))
        plt.subplot(121)
        plt.imshow(height_scaled_appearance_features)
        plt.subplot(122)
        plt.imshow(context_output[0, 0].detach().cpu().numpy())

    if best_up_sequence[0] == best_down_sequence[0]:
        same_both_directions = True
    else:
        same_both_directions = False

    if debug:
        return (
            vert_dicts,
            height_scaled_appearance_features,
            context_output[0, 0].detach().cpu().numpy(),
            vert_centre_features,
            sequence_predictions,
        )
    else:
        # remove keys that are not needed; 'appearance_guess', 'context_guess', 'context_features','appearance_features', 'visual_features'
        for vert_dict_idx, vert_dict in enumerate(vert_dicts):
            for key in list(vert_dict.keys()):
                if key not in ["predicted_label", "slice_nos", "polys","average_polygon"]:
                    del vert_dict[key]
            vert_dicts[vert_dict_idx] = vert_dict
        return vert_dicts


def conv_appearance_labelling(
    vert_dicts, appearance_net, softmax_temp=20, big_features=False
):
    appearance_net_device = next(appearance_net.parameters()).device
    volumes = [vert_dict["volume"] for vert_dict in vert_dicts]
    empty_vol_idxs = [
        idx for idx in range(len(vert_dicts)) if (vert_dicts[idx]["volume"].sum() == 0)
    ]
    appearance_net.eval()
    with torch.no_grad():
        appearance_predictions = []
        for volume in volumes:
            if big_features:
                appearance_predictions.append(
                    appearance_net.get_appearance_features(
                        torch.tensor(volume)
                        .unsqueeze(0)
                        .unsqueeze(0)
                        .float()
                        .to(appearance_net_device)
                    )
                )
            else:
                appearance_predictions.append(
                    appearance_net(
                        torch.tensor(volume)
                        .unsqueeze(0)
                        .unsqueeze(0)
                        .float()
                        .to(appearance_net_device)
                    )
                )
        visual_features = torch.cat(appearance_predictions, dim=0)
        appearance_predictions = torch.softmax(
            torch.cat(appearance_predictions, dim=0) / softmax_temp, dim=-1
        )

    for vert_dict_idx, vert_dict in enumerate(vert_dicts):
        vert_dict["appearance_features"] = appearance_predictions[vert_dict_idx]
        vert_dict["visual_features"] = visual_features[vert_dict_idx].cpu()

    for idx in empty_vol_idxs:
        vert_dicts[idx]["appearance_features"] = torch.zeros_like(
            vert_dicts[idx]["appearance_features"]
        ).float()
        vert_dicts[idx]["visual_features"] = (
            torch.zeros_like(vert_dicts[idx]["visual_features"]).float().cpu()
        )

    return vert_dicts


def construct_input_to_context_model(vert_dicts, scan, pixel_spacing):
    """
    Makes the image to be input into the context model.

    Inputs
    ------
    vert_dicts: list of dictionaries
        dictionaries containing information on each detected vertebra. Data structure output by detection pipeline.
    scan: np.array
        HxWxS size array containing the scan. H is height, W is width, S is sagittal slices
    pixel_spacing: float
        The distance between each pixel in mm

    Outputs
    -------
    height_scaled_appearance_features : np.array
        a HxW array containing the image used as input to the context array
    y_centroids: list
        a list of the y-coords of the centroids of each vertebra
    """

    appearance_features = [vert_dict["appearance_features"] for vert_dict in vert_dicts]
    box_coords = np.asarray([vert_dict["average_polygon"] for vert_dict in vert_dicts])
    y_centroids = (
        np.asarray([np.mean(box_coord[:, 1]) for box_coord in box_coords])
        * pixel_spacing[1]
    )
    y_maxes = (
        np.asarray([np.max(box_coord[:, 1]) for box_coord in box_coords])
        * pixel_spacing[1]
    )
    y_mins = (
        np.asarray([np.min(box_coord[:, 1]) for box_coord in box_coords])
        * pixel_spacing[1]
    )
    widths = (y_maxes - y_mins) / 2
    image_height = int(scan.shape[0] * pixel_spacing[0])
    height_scaled_appearance_features = np.zeros((int(image_height), 24))

    for i in range(len(vert_dicts)):
        height_scaled_appearance_features[
            int(y_centroids[i])
            - int(widths[i] / 2) : int(y_centroids[i])
            + int(widths[i] / 2),
            :,
        ] = (
            appearance_features[i].cpu().numpy()
        )

    return height_scaled_appearance_features, y_centroids


def beam_search_decoder(data, k, search_type="descending", penalise_skips=True):
    assert search_type in [
        "ascending",
        "descending",
    ], "Search type should be either 'ascending' or 'descending'"
    sequences = [[list(), 1.0]]
    # walk over each step in sequence
    for row in data:
        all_candidates = list()
        # expand each current candidate
        for i in range(len(sequences)):
            seq, score = sequences[i]
            for j in range(len(row)):
                candidate = [seq + [j], score - np.log(row[j])]
                if len(seq):
                    if search_type == "descending":
                        if j >= seq[-1] and j != 0:
                            candidate[1] += 1000
                        if j >= seq[-1] and j == 0:
                            candidate[1] += 100
                        if penalise_skips:
                            if np.abs(j - seq[-1]) > 1:
                                candidate[1] += 100

                    else:
                        if j <= seq[-1]:
                            candidate[1] += 1000
                        if penalise_skips:
                            if np.abs(j - seq[-1]) > 1:
                                candidate[1] += 100

                all_candidates.append(candidate)
        #  order all candidates by score
        ordered = sorted(all_candidates, key=lambda tup: tup[1])
        # select k best
        sequences = ordered[:k]

    return sequences


def two_way_beam_search(data, k, down_only=True, penalise_skips=False):
    down_sequences = beam_search_decoder(
        data, k, search_type="descending", penalise_skips=penalise_skips
    )
    if down_only:
        return down_sequences, True, down_sequences[0], down_sequences[0]
    data = list(reversed(data))
    reversed_up_sequences = beam_search_decoder(
        data, k, search_type="ascending", penalise_skips=penalise_skips
    )
    up_sequences = []
    for sequence in reversed_up_sequences:
        up_sequences.append([list(reversed(sequence[0])), sequence[1]])

    best_up_sequence = sorted(up_sequences, key=lambda tup: tup[1])[0]
    best_down_sequence = sorted(down_sequences, key=lambda tup: tup[1])[0]

    if best_up_sequence == best_down_sequence:
        same_both_directions = True
    else:
        same_both_directions = False

    all_sequences = up_sequences + down_sequences
    all_sequences = sorted(all_sequences, key=lambda tup: tup[1])
